"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SqlDDLBuilderAdapter = void 0;
/* eslint-disable @typescript-eslint/no-unused-vars */
const lambdaorm_base_1 = require("lambdaorm-base");
const domain_1 = require("../../../../query/domain");
const DDLBuilderAdapter_1 = require("../base/DDLBuilderAdapter");
class SqlDDLBuilderAdapter extends DDLBuilderAdapter_1.DDLBuilderAdapter {
    truncateEntity(entity) {
        let text = this.dialect.ddl('truncateEntity');
        text = text.replace('{name}', this.dialect.delimiter(entity.mapping));
        return new domain_1.Query({ action: lambdaorm_base_1.SentenceAction.truncateEntity, dialect: this.source.dialect, source: this.source.name, sentence: text, entity: entity.name });
    }
    createEntity(entity) {
        const define = [];
        for (const i in entity.properties) {
            const property = entity.properties[i];
            if (!property.view) {
                define.push(this.createColumn(entity, property));
            }
        }
        if (entity.primaryKey && entity.primaryKey.length > 0) {
            define.push(this.createPk(entity, entity.primaryKey));
        }
        let text = this.dialect.ddl('createEntity');
        text = text.replace('{name}', this.dialect.delimiter(entity.mapping));
        text = text.replace('{define}', define.join(','));
        return new domain_1.Query({ action: lambdaorm_base_1.SentenceAction.createEntity, dialect: this.source.dialect, source: this.source.name, sentence: text, entity: entity.name });
    }
    createColumn(entity, property) {
        let type = this.dialect.type(property.type);
        if (type === undefined) {
            throw new lambdaorm_base_1.SchemaError(`Undefined type for ${entity.name}.${property.name}`);
        }
        type = property.length ? type.replace('{0}', property.length.toString()) : type;
        const required = property.required ? this.dialect.other('notNullable') : '';
        let text = property.autoIncrement ? this.dialect.ddl('incrementalColumDefine') : this.dialect.ddl('columnDefine');
        text = text.replace('{name}', this.dialect.delimiter(property.mapping));
        text = text.replace('{type}', type);
        text = text.replace('{required}', required);
        return text;
    }
    createPk(entity, primaryKey) {
        const columns = [];
        const columnTemplate = this.dialect.other('column');
        for (const primaryKeyItem of primaryKey) {
            const property = entity.properties.find(p => p.name === primaryKeyItem);
            columns.push(columnTemplate.replace('{name}', this.dialect.delimiter(property.mapping)));
        }
        let text = this.dialect.ddl('createPk');
        text = text.replace('{name}', this.dialect.delimiter(entity.mapping + '_PK'));
        text = text.replace('{columns}', columns.join(','));
        return text;
    }
    createFk(entity, relation) {
        const column = entity.properties.find(p => p.name === relation.from);
        const fEntity = this.mapping.getEntity(relation.entity);
        const fColumn = fEntity.properties.find(p => p.name === relation.to);
        const alterEntity = this.dialect.ddl('alterTable').replace('{name}', this.dialect.delimiter(entity.mapping));
        let text = this.dialect.ddl('createFk');
        text = text.replace('{name}', this.dialect.delimiter(entity.mapping + '_' + relation.name + '_FK'));
        text = text.replace('{column}', this.dialect.delimiter(column.mapping));
        text = text.replace('{fTable}', this.dialect.delimiter(fEntity.mapping));
        text = text.replace('{fColumn}', this.dialect.delimiter(fColumn.mapping));
        return new domain_1.Query({ action: lambdaorm_base_1.SentenceAction.addFk, dialect: this.source.dialect, source: this.source.name, sentence: alterEntity + ' ' + text, entity: entity.name });
    }
    createSequence(entity) {
        let text = this.dialect.ddl('createSequence');
        text = text.replace('{name}', this.dialect.delimiter(entity.sequence));
        return new domain_1.Query({ action: lambdaorm_base_1.SentenceAction.createSequence, dialect: this.source.dialect, source: this.source.name, sentence: text, entity: entity.name });
    }
    createIndex(entity, index) {
        const columns = [];
        const columnTemplate = this.dialect.other('column');
        for (const field of index.fields) {
            const propertyMapping = entity.properties.find(p => p.name === field);
            if (propertyMapping) {
                columns.push(columnTemplate.replace('{name}', this.dialect.delimiter(propertyMapping.mapping)));
            }
        }
        let text = this.dialect.ddl('createIndex');
        text = text.replace('{name}', this.dialect.delimiter(entity.mapping + '_' + index.name));
        text = text.replace('{table}', this.dialect.delimiter(entity.mapping));
        text = text.replace('{columns}', columns.join(','));
        return new domain_1.Query({ action: lambdaorm_base_1.SentenceAction.createIndex, dialect: this.source.dialect, source: this.source.name, sentence: text, entity: entity.name });
    }
    alterProperty(entity, property) {
        let text = this.property(entity, property);
        text = this.dialect.ddl('alterProperty').replace('{columnDefine}', text);
        const alterEntity = this.dialect.ddl('alterTable').replace('{name}', this.dialect.delimiter(entity.mapping));
        return new domain_1.Query({ action: lambdaorm_base_1.SentenceAction.alterProperty, dialect: this.source.dialect, source: this.source.name, sentence: alterEntity + ' ' + text, entity: entity.name });
    }
    addProperty(entity, property) {
        let text = this.property(entity, property);
        text = this.dialect.ddl('addProperty').replace('{columnDefine}', text);
        const alterEntity = this.dialect.ddl('alterTable').replace('{name}', this.dialect.delimiter(entity.mapping));
        return new domain_1.Query({ action: lambdaorm_base_1.SentenceAction.addProperty, dialect: this.source.dialect, source: this.source.name, sentence: alterEntity + ' ' + text, entity: entity.name });
    }
    property(entity, property) {
        const propertyMapping = this.mapping.getProperty(entity.name, property.name);
        let type = this.dialect.type(property.type);
        if (type === undefined) {
            throw new lambdaorm_base_1.SchemaError(`Undefined type for ${entity.name}.${property.name}`);
        }
        type = property.length ? type.replace('{0}', property.length.toString()) : type;
        const required = property.required ? this.dialect.other('notNullable') : '';
        let text = property.autoIncrement ? this.dialect.ddl('incrementalColumDefine') : this.dialect.ddl('columnDefine');
        text = text.replace('{name}', this.dialect.delimiter(propertyMapping.mapping));
        text = text.replace('{type}', type);
        return text.replace('{required}', required);
    }
    addPk(entity, primaryKeys) {
        return this.addKeys(entity, lambdaorm_base_1.SentenceAction.addPk, '_PK', primaryKeys);
    }
    addUk(entity, uniqueKeys) {
        return this.addKeys(entity, lambdaorm_base_1.SentenceAction.addUk, '_UK', uniqueKeys);
    }
    addKeys(entity, ddl, suffix, keys) {
        const columns = [];
        const columnTemplate = this.dialect.other('column');
        for (const key of keys) {
            const property = entity.properties.find(p => p.name === key);
            if (property) {
                columns.push(columnTemplate.replace('{name}', this.dialect.delimiter(property.mapping)));
            }
        }
        const alterEntity = this.dialect.ddl('alterTable').replace('{name}', this.dialect.delimiter(entity.mapping));
        let text = this.dialect.ddl(ddl);
        text = text.replace('{name}', this.dialect.delimiter(entity.mapping + suffix));
        text = text.replace('{columns}', columns.join(','));
        return new domain_1.Query({ action: ddl, dialect: this.source.dialect, source: this.source.name, sentence: alterEntity + ' ' + text, entity: entity.name });
    }
    addFk(entity, relation) {
        const column = entity.properties.find(p => p.name === relation.from);
        if (!column) {
            throw new lambdaorm_base_1.SchemaError(`Property ${relation.from} not found in entity ${entity.name}`);
        }
        const fEntity = this.mapping.getEntity(relation.entity);
        if (!fEntity) {
            throw new lambdaorm_base_1.SchemaError(`Entity ${relation.entity} not found`);
        }
        const fColumn = fEntity.properties.find(p => p.name === relation.to);
        if (!fColumn) {
            throw new lambdaorm_base_1.SchemaError(`Property ${relation.to} not found in entity ${fEntity.name}`);
        }
        const alterEntity = this.dialect.ddl('alterTable').replace('{name}', this.dialect.delimiter(entity.mapping));
        let text = this.dialect.ddl(lambdaorm_base_1.SentenceAction.addFk);
        text = text.replace('{name}', this.dialect.delimiter(entity.mapping + '_' + relation.name + '_FK'));
        text = text.replace('{column}', this.dialect.delimiter(column.mapping));
        text = text.replace('{fTable}', this.dialect.delimiter(fEntity.mapping));
        text = text.replace('{fColumn}', this.dialect.delimiter(fColumn.mapping));
        return new domain_1.Query({ action: lambdaorm_base_1.SentenceAction.addFk, dialect: this.source.dialect, source: this.source.name, sentence: alterEntity + ' ' + text, entity: entity.name });
    }
    dropEntity(entity) {
        let text = this.dialect.ddl(lambdaorm_base_1.SentenceAction.dropEntity);
        text = text.replace('{name}', this.dialect.delimiter(entity.mapping));
        return new domain_1.Query({ action: lambdaorm_base_1.SentenceAction.dropEntity, dialect: this.source.dialect, source: this.source.name, sentence: text, entity: entity.name });
    }
    dropProperty(entity, property) {
        const propertyMapping = this.mapping.getProperty(entity.name, property.name);
        const alterEntity = this.dialect.ddl('alterTable').replace('{name}', this.dialect.delimiter(entity.mapping));
        let text = this.dialect.ddl('dropProperty');
        text = text.replace('{name}', this.dialect.delimiter(propertyMapping.mapping));
        return new domain_1.Query({ action: lambdaorm_base_1.SentenceAction.dropProperty, dialect: this.source.dialect, source: this.source.name, sentence: alterEntity + ' ' + text, entity: entity.name });
    }
    dropPk(entity) {
        const alterEntity = this.dialect.ddl('alterTable').replace('{name}', this.dialect.delimiter(entity.mapping));
        let text = this.dialect.ddl(lambdaorm_base_1.SentenceAction.dropPk);
        text = text.replace('{name}', this.dialect.delimiter(entity.mapping + '_PK'));
        return new domain_1.Query({ action: lambdaorm_base_1.SentenceAction.dropPk, dialect: this.source.dialect, source: this.source.name, sentence: alterEntity + ' ' + text, entity: entity.name });
    }
    dropUk(entity) {
        const alterEntity = this.dialect.ddl('alterTable').replace('{name}', this.dialect.delimiter(entity.mapping));
        let text = this.dialect.ddl(lambdaorm_base_1.SentenceAction.dropUk);
        text = text.replace('{name}', this.dialect.delimiter(entity.mapping + '_UK'));
        return new domain_1.Query({ action: lambdaorm_base_1.SentenceAction.dropUk, dialect: this.source.dialect, source: this.source.name, sentence: alterEntity + ' ' + text, entity: entity.name });
    }
    setNull(entity, relation) {
        const alias = 'a';
        const templateColumn = this.dialect.other('column');
        const propertyFrom = entity.properties.find(p => p.name === relation.from);
        if (!propertyFrom) {
            throw new lambdaorm_base_1.SchemaError(`not found relation form ${entity.name}.${relation.name}.${relation.from} `);
        }
        const column = templateColumn.replace('{name}', propertyFrom.mapping);
        const templateAssign = this.dialect.operator('=', 2);
        let assign = templateAssign.replace('{0}', column);
        const _null = this.dialect.other('null');
        assign = assign.replace('{1}', _null);
        let text = this.dialect.dml(lambdaorm_base_1.SentenceAction.update);
        text = text.replace('{name}', this.dialect.delimiter(entity.mapping));
        text = this.helper.str.replace(text, '{alias}', alias);
        text = text.replace('{assigns}', assign);
        return new domain_1.Query({ action: lambdaorm_base_1.SentenceAction.update, dialect: this.source.dialect, source: this.source.name, sentence: text, entity: entity.name });
    }
    dropFk(entity, relation) {
        const alterEntity = this.dialect.ddl('alterTable').replace('{name}', this.dialect.delimiter(entity.mapping));
        let text = this.dialect.ddl(lambdaorm_base_1.SentenceAction.dropFk);
        text = text.replace('{name}', this.dialect.delimiter(entity.mapping + '_' + relation.name + '_FK'));
        return new domain_1.Query({ action: lambdaorm_base_1.SentenceAction.dropFk, dialect: this.source.dialect, source: this.source.name, sentence: alterEntity + ' ' + text, entity: entity.name });
    }
    dropIndex(entity, index) {
        let text = this.dialect.ddl(lambdaorm_base_1.SentenceAction.dropIndex);
        text = text.replace('{name}', this.dialect.delimiter(entity.mapping + '_' + index.name));
        text = text.replace('{table}', this.dialect.delimiter(entity.mapping));
        return new domain_1.Query({ action: lambdaorm_base_1.SentenceAction.dropIndex, dialect: this.source.dialect, source: this.source.name, sentence: text, entity: entity.name });
    }
    dropSequence(entity) {
        let text = this.dialect.ddl(lambdaorm_base_1.SentenceAction.dropSequence);
        text = text.replace('{name}', this.dialect.delimiter(entity.sequence));
        return new domain_1.Query({ action: lambdaorm_base_1.SentenceAction.dropSequence, dialect: this.source.dialect, source: this.source.name, sentence: text, entity: entity.name });
    }
}
exports.SqlDDLBuilderAdapter = SqlDDLBuilderAdapter;
//# sourceMappingURL=ddlBuilder.js.map