"use strict";
/* eslint-disable @typescript-eslint/ban-types */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.orm = exports.Orm = void 0;
const application_1 = require("../../shared/application");
const h3lp_1 = require("h3lp");
const lambdaorm_base_1 = require("lambdaorm-base");
const infrastructure_1 = require("../../connection/infrastructure");
const expressionsBuilder_1 = require("./expressionsBuilder");
const _3xpr_1 = require("3xpr");
const infrastructure_2 = require("../../operand/infrastructure");
const facadeBuilder_1 = require("../../sentence/infrastructure/facadeBuilder");
const infrastructure_3 = require("../../expressions/infrastructure");
const infrastructure_ts_1 = require("../../execution/infrastructure.ts");
const infrastructure_4 = require("../../stage/infrastructure");
const infrastructure_5 = require("../../sentence/infrastructure");
const application_2 = require("../../execution/application");
const ormLibrary_1 = require("./ormLibrary");
/**
 * Facade through which you can access all the functionalities of the library.
 */
class Orm {
    constructor(workspace = process.cwd()) {
        this.expressions = new expressionsBuilder_1.OrmExpressionsBuilder().build();
        new ormLibrary_1.OrmLibrary(this).load();
        this.helper = new application_1.Helper(new _3xpr_1.OperandHelper(this.expressions.constBuilder), h3lp_1.h3lp);
        this.language = new infrastructure_5.SentenceLanguageServiceBuilder(this.helper).build();
        this.connection = new infrastructure_1.ConnectionFacadeBuilder(this.helper).build();
        this.schema = new lambdaorm_base_1.SchemaFacadeBuilder(this.expressions, this.helper).build(workspace);
        this.executor = new infrastructure_ts_1.ExecutorBuilder(this.connection, this.language, this.expressions, this.helper).build(this.schema);
        this.operand = new infrastructure_2.OperandFacadeBuilder(this.expressions, this.helper).build(this.schema);
        this.sentence = new facadeBuilder_1.SentenceFacadeBuilder(this.expressions, this.helper).build(this.schema, this.operand);
        this.expression = new infrastructure_3.ExpressionFacadeBuilder(this.language, this.executor, this.expressions, this.helper).build(this.sentence, this.schema);
        this.stage = new infrastructure_4.StageFacadeBuilder(this.language, this.executor, this.helper).build(this.schema, this.expression);
    }
    /**
  * Singleton
  */
    static get instance() {
        if (!this._instance) {
            this._instance = new Orm();
        }
        return this._instance;
    }
    get defaultStage() {
        return this.schema.stage.get();
    }
    /**
 * initialize the orm library
 * @param source optional parameter to specify the location of the configuration file. In the case that it is not passed, it is assumed that it is "lambdaORM.yaml" in the root of the project
 * @returns promise void
 */
    init(source, connect = true) {
        var _a, _b, _c;
        return __awaiter(this, void 0, void 0, function* () {
            const schema = yield this.schema.initialize(source || this.schema.workspace);
            // set connections
            if (connect && ((_a = schema.infrastructure) === null || _a === void 0 ? void 0 : _a.sources)) {
                for (const source of schema.infrastructure.sources.filter(p => this.helper.val.isNotEmpty(p.connection))) {
                    this.connection.load(source);
                }
            }
            // add enums
            if (schema.domain.enums) {
                for (const _enum of schema.domain.enums) {
                    const values = [];
                    if (_enum.values) {
                        for (const enumValue of _enum.values) {
                            values.push([enumValue.name, enumValue.value]);
                        }
                    }
                    this.expressions.addEnum(_enum.name, values);
                }
            }
            // start
            if ((_b = schema.application) === null || _b === void 0 ? void 0 : _b.start) {
                for (const task of schema.application.start) {
                    if (task.condition === undefined || this.expressions.eval(task.condition)) {
                        yield this.expressions.evalAsync(task.expression);
                    }
                }
            }
            // add listeners
            if ((_c = schema.application) === null || _c === void 0 ? void 0 : _c.listeners) {
                for (const listener of schema.application.listeners) {
                    const observer = new application_2.ExecutionActionObserver(listener, this.expressions);
                    this.subscribe(observer);
                }
            }
            return schema;
        });
    }
    /**
  * Frees the resources used, for example the connection pools
  */
    end() {
        var _a;
        return __awaiter(this, void 0, void 0, function* () {
            // ends task
            const schema = this.schema.schema;
            if ((_a = schema.application) === null || _a === void 0 ? void 0 : _a.end) {
                for (const task of schema.application.end) {
                    if (task.condition === undefined || this.expressions.eval(task.condition)) {
                        yield this.expressions.evalAsync(task.expression);
                    }
                }
            }
            yield this.connection.end();
        });
    }
    /**
     * Get workspace path
     */
    get workspace() {
        return this.schema.workspace;
    }
    /**
     * Get dialect of source
     * @param source Name of source
     * @returns
     */
    dialect(source) {
        return this.schema.source.get(source).dialect;
    }
    normalize(expression) {
        const _expression = this.toExpression(expression);
        return this.operand.normalize(_expression);
    }
    model(expression) {
        const _expression = this.toExpression(expression);
        return this.sentence.model(_expression);
    }
    parameters(expression) {
        const _expression = this.toExpression(expression);
        return this.sentence.parameters(_expression);
    }
    constraints(expression) {
        const _expression = this.toExpression(expression);
        return this.sentence.constraints(_expression);
    }
    metadata(expression) {
        const _expression = this.toExpression(expression);
        return this.sentence.metadata(_expression);
    }
    plan(expression, options) {
        const _expression = this.toExpression(expression);
        const _options = options !== undefined && typeof options === 'string' ? JSON.parse(options) : options || {};
        return this.expression.plan(_expression, _options);
    }
    execute(expression, data = {}, options) {
        return __awaiter(this, void 0, void 0, function* () {
            if (expression === undefined || expression === null) {
                throw new Error('expression is empty');
            }
            const _expression = this.toExpression(expression);
            if (_expression === '') {
                throw new Error('expression is empty');
            }
            const _data = data !== undefined && typeof data === 'string' ? JSON.parse(data) : data || {};
            const _options = options !== undefined && typeof options === 'string' ? JSON.parse(options) : options || {};
            return this.expression.execute(_expression, _data, _options);
        });
    }
    /**
     * Create a transaction
     * @param options options of execution
     * @param callback Code to be executed in transaction
     */
    transaction(options, callback) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.expression.transaction(options, callback);
        });
    }
    toExpression(expression) {
        return typeof expression !== 'string' ? this.expressions.convert(expression, 'function')[0] : expression;
    }
    subscribe(observer) {
        this.executor.subscribe(observer);
    }
    unsubscribe(observer) {
        this.executor.unsubscribe(observer);
    }
}
exports.Orm = Orm;
exports.orm = Orm.instance;
//# sourceMappingURL=orm.js.map